//
/*
	The brains of othello playing program.
	TheMaster Version 2.0

	Version 2.0 improvements over 1.1:
		- Quiessence search
		- Mobility and Potential Mobility considerstion
		- order of moves considered more logical to increase speed

	CopyRight(c) Gaurang Khetan, Pune.
	Date: 17/4/1999
*/
//
#include <MEM.H>
#include <STDIO.H>
#include <ASSERT.H>
#include <CONIO.H>
#include <STDLIB.H>

typedef long signed int score;
typedef signed int id;
typedef signed int movetype;
typedef signed int pos;
enum boolean { false, true };

#include "BRAIN20.H"  // contains class definition of TheMaster2

//class alloc_class member functions
boolean alloc_class::set()
	{
	used_board = used_move_struc = 0;
	start_board = new pos[sizeof(pos)*64*40]; // space for boards
	start_move_struc = new char[35*20*40]; // space for move structures

	if ( (start_board==0) || (start_move_struc==0) )
		return false;
	else
		return true;
	}

void alloc_class::reset()
	{
	delete start_board;
	delete start_move_struc;
	}

pos* alloc_class::new_board()
	{
	assert(used_board<39);

	used_board++;
	return start_board + ( (used_board-1) * sizeof(pos) * 64 );
	}

void alloc_class::delete_board(pos* p)
	{
	assert (used_board>0);
	assert (p == ( start_board + ( (used_board-1) * sizeof(pos) * 64) ) );

	used_board--;
	}

char* alloc_class::new_move_struc()
	{
	assert(used_move_struc<39);
	used_move_struc++;
	return start_move_struc + ( (used_move_struc-1) * 35 * 20 );
	}

void alloc_class::delete_move_struc(char* p)
	{
	assert (used_move_struc>0);
	assert (p == ( start_move_struc + ( (used_move_struc-1) * 35 * 20) ) );

	used_move_struc--;
	}

//class brain member functions
TheMaster2::TheMaster2()
	{
	eval_level = 3; min_depth = 8;
	move = -1;
	start_alpha = -999999; start_beta = 999999;

	//make order of moves to be considered from outwards spiralling inwards
	order[0] = 0; order[1] = 7; order[2] = 56; order[3] = 63;
	order[4] = 2; order[5] = 3; order[6] = 4; order[7] = 5;
	order[8] = 23; order[9] = 31; order[10] = 39; order[11] = 47;
	order[12] = 16; order[13] = 24; order[14] = 32; order[15] = 40;
	order[16] = 58; order[17] = 59; order[18] = 60; order[19] = 61;
	order[20] = 18; order[21] = 21; order[22] = 42; order[23] = 45;
	order[24] = 43; order[25] = 37; order[26] = 20; order[27] = 26;
	order[28] = 34; order[29] = 44; order[30] = 29; order[31] = 19;
	order[32] = 10; order[33] = 11; order[34] = 12; order[35] = 13;
	order[36] = 22; order[37] = 30; order[38] = 38; order[39] = 46;
	order[40] = 17; order[41] = 25; order[42] = 33; order[43] = 41;
	order[44] = 50; order[45] = 51; order[46] = 52; order[47] = 53;
	order[48] = 48; order[49] = 55; order[50] = 15; order[51] = 8;
	order[52] = 57; order[53] = 62; order[54] = 6; order[55] = 1;
	order[56] = 49; order[57] = 54; order[58] = 14; order[59] = 9;

	if (!(alloc.set()))
		{
		printf("Out of Memory for making the computer play! ");
		getch();
		exit(1);
		}
	}

TheMaster2::~TheMaster2()
	{
	alloc.reset();
	}

void TheMaster2::set_id(id meid, id youid)
	{ me_id = meid; you_id = youid; }

void TheMaster2::set_depth(int md)
	{ min_depth = md; }

void TheMaster2::set_eval_level(int t)
	{ eval_level = t; }

score TheMaster2::get_cur_score()
	{ return cur_score; }

boolean TheMaster2::is_legal(id who, movetype m)
	{
	if (board[m]) return false;

	int r = m/8, c = m%8;

	if (c<=5)
		if (board[m+1]==opp(who))
			for ( int i=m+2; i%8!=0; i++)
				if (board[i]==who) return true;
				else if (board[i]==0) break;
	if (c>=2)
		if (board[m-1]==opp(who))
			for ( int i=m-2; i%8!=7 && i!=-1; i--)
				if (board[i]==who) return true;
				else if (board[i]==0) break;
	if (r>=2)
		if (board[m-8]==opp(who))
			for ( int i=m-16; i>=0 ; i-=8)
				if (board[i]==who) return true;
				else if (board[i]==0) break;
	if (r<=5)
		if (board[m+8]==opp(who))
			for ( int i=m+16; i<=63 ; i+=8)
				if (board[i]==who) return true;
				else if (board[i]==0) break;
	if (r<=5 && c<=5)
		if (board[m+9]==opp(who))
			for ( int i=m+18; (i<=63) && ((i%8)!=0) ; i+=9)
				if (board[i]==who) return true;
				else if (board[i]==0) break;
	if (r<=5 && c>=2)
		if (board[m+7]==opp(who))
			for ( int i=m+14; (i<=63) && ((i%8)!=7) ; i+=7)
				if (board[i]==who) return true;
				else if (board[i]==0) break;
	if (r>=2 && c<=5)
		if (board[m-7]==opp(who))
			for ( int i=m-14; i>=0 && i%8 != 0 ; i-=7)
				if (board[i]==who) return true;
				else if (board[i]==0) break;
	if (r>=2 && c>=2)
		if (board[m-9]==opp(who))
			for ( int i=m-18; (i>=0) && ((i%8)!=7) ; i-=9)
				if (board[i]==who) return true;
				else if (board[i]==0) break;

	return false;
	}

void TheMaster2::play_move(id who, movetype m )
	{
	board[m] = who;

	int r = m/8, c = m%8;
	id other = me_id + you_id - who;

	if (c<=5)
		if (board[m+1]==other)
			for ( int i=m+2; i%8!=0; i++)
				{
				if (board[i]==other) continue;
				if (board[i]==who)
					for (int j=m+1; j!=i; j++)
						board[j] = who;
				break;
				}
	if (c>=2)
		if (board[m-1]==opp(who))
			for ( int i=m-2; i%8!=7 && i!=-1; i--)
				{
				if (board[i]==other) continue;
				if (board[i]==who)
					for (int j=m-1; j!=i; j--)
						board[j] = who;
				break;
				}
	if (r>=2)
		if (board[m-8]==opp(who))
			for ( int i=m-16; i>=0 ; i-=8)
				{
				if (board[i]==other) continue;
				if (board[i]==who)
					for (int j=m-8; j!=i; j-=8)
						board[j] = who;
				break;
				}
	if (r<=5)
		if (board[m+8]==opp(who))
			for ( int i=m+16; i<=63 ; i+=8)
				{
				if (board[i]==other) continue;
				if (board[i]==who)
					for (int j=m+8; j!=i; j+=8)
						board[j] = who;
				break;
				}
	if (r<=5 && c<=5)
		if (board[m+9]==opp(who))
			for ( int i=m+18; (i<=63) && ((i%8)!=0) ; i+=9)
				{
				if (board[i]==other) continue;
				if (board[i]==who)
					for (int j=m+9; j!=i; j+=9)
						board[j] = who;
				break;
				}
	if (r<=5 && c>=2)
		if (board[m+7]==opp(who))
			for ( int i=m+14; (i<=63) && ((i%8)!=7) ; i+=7)
				{
				if (board[i]==other) continue;
				if (board[i]==who)
					for (int j=m+7; j!=i; j+=7)
						board[j] = who;
				break;
				}
	if (r>=2 && c<=5)
		if (board[m-7]==opp(who))
			for ( int i=m-14; i>=0 && i%8 != 0 ; i-=7)
				{
				if (board[i]==other) continue;
				if (board[i]==who)
					for (int j=m-7; j!=i; j-=7)
						board[j] = who;
				break;
				}
	if (r>=2 && c>=2)
		if (board[m-9]==opp(who))
			for ( int i=m-18; (i>=0) && ((i%8)!=7) ; i-=9)
				{
				if (board[i]==other) continue;
				if (board[i]==who)
					for (int j=m-9; j!=i; j-=9)
						board[j] = who;
				break;
				}
	}

movetype TheMaster2::next_move(id plid, int& k)
	{
	if (k==60) return -1;

	for (;k<=59; k++)
		if (is_legal(plid, order[k]))
			{
			k = k+1;
			return order[k-1];
			}

	return -1;
	}

int TheMaster2::no_of_legal_moves(id who)
	{
	int to_return=0;
	for(int i=0; i<=63; i++)
		if ( is_legal(who, i) ) to_return++;

	return to_return;
	}


void TheMaster2::set_up_legal_moves(id who, char* move_struc)
	{

	int move_no = 0;
	id other = you_id + me_id - who;

	for (int z = 0; z<=60; z++)
		{
		movetype m = order[z];

		move_struc[move_no*20] = m;	//put m on it for trial
		int move_effect = 1;  //it will be overwritten next loop if it
									// does not present itself as legal move

		//begin testing
		int r = m/8, c = m%8;

		if (board[m]) continue;

		if (c<=5)
			if (board[m+1]==other)
				for ( int i=m+2; i%8!=0; i++)
					{
					if (board[i]==other) continue;
					if (board[i]==who)
						for (int j=m+1; j!=i; j++)
							{
							move_struc[move_no*20 + move_effect] = j;
							move_effect++;
							}
					break;
					}
		if (c>=2)
			if (board[m-1]==other)
				for ( int i=m-2; i%8!=7 && i!=-1; i--)
					{
					if (board[i]==other) continue;
					if (board[i]==who)
						for (int j=m-1; j!=i; j--)
							{
							move_struc[move_no*20 + move_effect] = j;
							move_effect++;
							}
					break;
					}
		if (r>=2)
			if (board[m-8]==other)
				for ( int i=m-16; i>=0 ; i-=8)
					{
					if (board[i]==other) continue;
					if (board[i]==who)
						for (int j=m-8; j!=i; j-=8)
							{
							move_struc[move_no*20 + move_effect] = j;
							move_effect++;
							}
					break;
					}
		if (r<=5)
			if (board[m+8]==other)
				for ( int i=m+16; i<=63 ; i+=8)
					{
					if (board[i]==other) continue;
					if (board[i]==who)
						for (int j=m+8; j!=i; j+=8)
							{
							move_struc[move_no*20 + move_effect] = j;
							move_effect++;
							}
					break;
					}
		if (r<=5 && c<=5)
			if (board[m+9]==other)
				for ( int i=m+18; (i<=63) && ((i%8)!=0) ; i+=9)
					{
					if (board[i]==other) continue;
					if (board[i]==who)
						for (int j=m+9; j!=i; j+=9)
							{
							move_struc[move_no*20 + move_effect] = j;
							move_effect++;
							}
					break;
					}
		if (r<=5 && c>=2)
			if (board[m+7]==other)
				for ( int i=m+14; (i<=63) && ((i%8)!=7) ; i+=7)
					{
					if (board[i]==other) continue;
					if (board[i]==who)
						for (int j=m+7; j!=i; j+=7)
							{
							move_struc[move_no*20 + move_effect] = j;
							move_effect++;
							}
					break;
					}
		if (r>=2 && c<=5)
			if (board[m-7]==other)
				for ( int i=m-14; i>=0 && i%8 != 0 ; i-=7)
					{
					if (board[i]==other) continue;
					if (board[i]==who)
						for (int j=m-7; j!=i; j-=7)
							{
							move_struc[move_no*20 + move_effect] = j;
							move_effect++;
							}
					break;
					}
		if (r>=2 && c>=2)
			if (board[m-9]==other)
				for ( int i=m-18; (i>=0) && ((i%8)!=7) ; i-=9)
					{
					if (board[i]==other) continue;
					if (board[i]==who)
						for (int j=m-9; j!=i; j-=9)
							{
							move_struc[move_no*20 + move_effect] = j;
							move_effect++;
							}
					break;
					}

		if (move_effect>1)
			{
			move_struc[move_no*20 + move_effect] = 100;	//indicate termination
			move_no++;	//set pointer ready for next legal move
			}

		}

	move_struc[move_no*20] = 100;
	assert(move_no<34);

	}

boolean TheMaster2::is_legal(char* move_struc, movetype m)
	{
	int i = 0;

	while(1)
		{
		int t = move_struc[i*20];
		if (t==100) return false;
		if (t==m) return true;
		i++;
		assert ( i<=34);
		}
	}

int TheMaster2::no_of_legal_moves(char* move_struc)
	{
	for(int i=0; i<=34; i++)
		if (move_struc[i*20]==100) break;
	return i;
	}

void TheMaster2::play_move(id who, char* move_struc, int move_count)
	{
	int t;
	for (int i=0; i<=19; i++)
		{
		t = move_struc[move_count*20 + i];
		if (t==100) break;
		board[t] = who;
		}
	assert(t==100);
	}

movetype TheMaster2::next_move(char* move_struc, int& k)
	{
	int t = move_struc[k*20];

	if (t==100) return -1;	//moves exhausted

	//move available
	k++;         //set up k for the next time
	return t;	//return the just read move

	}

int TheMaster2::moves_finished()
	{
	int to_return=60;

	for(int i=0; i<=63; i++)
		if (board[i]==0) to_return--;

	return to_return;
	}

int TheMaster2::no_of_pieces(id who)
	{
	int to_return=0;
	for (int i=0; i<=63; i++)
		to_return += ( (board[i]==who)? 1 : 0 );
	return to_return;
	}

movetype TheMaster2::make_move(pos* brd)
	{
	board = brd;

	moves_over = moves_finished();
	if (moves_over>=48)
		cur_score = eg_evaluate(me_id, start_alpha, start_beta);
	else
		cur_score = evaluate(me_id, start_alpha, start_beta, min_depth);


	return move;
	}

score TheMaster2::evaluate(id plid, score alpha, score beta, int height)
	{
	score v,e;
	boolean done;
	movetype p=-1, best_move=-1;
	int move_count=0;
	char* move_struc;
	int num_moves;

	move_struc = alloc.new_move_struc();
	set_up_legal_moves ( plid, move_struc );

	num_moves = no_of_legal_moves (move_struc);

	if (height == 0)
		{
		/*if (  				// factors indicating non-quiet conditions
			(num_moves==1)
			is_legal(move_struc, 0) || is_legal(move_struc, 7)
			|| is_legal(move_struc, 56) || is_legal(move_struc, 63)
			)
			{
			height = 1;
			}
		else
			{ */
			v = stat_eval(plid);
			if (plid==you_id) v = -v;  //my imagination cannot reach there
			move = best_move;

			alloc.delete_move_struc(move_struc);
			return v;
		 /*	}*/
		}


	if (num_moves == 0)
		{
		//find opponent's legal moves
		/*char* temp_move_struc = alloc.new_move_struc();
		set_up_legal_moves ( opp(plid), temp_move_struc);
		int t = no_of_legal_moves(temp_move_struc);
		alloc.delete_move_struc(temp_move_struc);*/
		int q = no_of_legal_moves(opp(plid));

		if (q==0)
			{
			alloc.delete_move_struc(move_struc);
			return stat_eval(plid);	//game has already ended
			}

		pos* t = board;
		board = alloc.new_board();
		memcpy(board, t, 64*sizeof(pos));

		v = -evaluate(opp(plid), -beta, -alpha, height-1);

		alloc.delete_board(board);
		board = t;
		}
	else
		{
		v = alpha;
		do
			{
			done = false;
			p = next_move(move_struc, move_count);
			if (p != -1)
				{
				pos* t = board;
				board = alloc.new_board();

				memcpy(board, t, 64*sizeof(pos));

				play_move(plid, move_struc, move_count-1);
				moves_over++;

				e = -evaluate(opp(plid), -beta, -v, height-1);

				if (e > v) { v = e; best_move = p; }
				if (v >= beta) done = true;

				alloc.delete_board(board);
				board = t;

				moves_over--;
				}
			}
		while ( p!= -1 && !done);
		}

	move = best_move;

	alloc.delete_move_struc(move_struc);

	return v;
	}

score TheMaster2::eg_evaluate(id plid, score alpha, score beta)
	{
	score v,e;
	boolean done;
	movetype p=-1, best_move=-1;
	int move_count=0;
	char* move_struc;

	move_struc = alloc.new_move_struc();

	set_up_legal_moves ( plid, move_struc );

	if (no_of_legal_moves(move_struc) == 0)
		{
		int q = no_of_legal_moves(opp(plid));

		if (q==0)
			{
			alloc.delete_move_struc(move_struc);
			return no_of_pieces(plid) - no_of_pieces(opp(plid));
			}

		pos* t = board;
		board = alloc.new_board();
		memcpy(board, t, 64*sizeof(pos));

		v = -eg_evaluate(opp(plid), -beta, -alpha);

		alloc.delete_board(board);
		board = t;
		}
	else
		{
		v = alpha;
		do
			{
			done = false;
			p = next_move(move_struc, move_count);
			if (p != -1)
				{
				pos* t = board;
				board = alloc.new_board();

				memcpy(board, t, 64*sizeof(pos));

				play_move(plid, move_struc, move_count-1);

				e = -eg_evaluate(opp(plid), -beta, -v);

				if (e > v) { v = e; best_move = p; }
				if (v >= beta) done = true;

				alloc.delete_board(board);
				board = t;

				}
			}
		while ( p!= -1 && !done);
		}

	move = best_move;

	alloc.delete_move_struc(move_struc);

	return v;
	}

score TheMaster2::stat_eval(id who)
	{
	score sum, sum1=0, sum2=0, sum3=0, sum4=0, sum5=0;
	float a,b,c,d,e;

	if      (moves_over < 20) { a = 9; b = 2; c = 2; d = 9; e = 6; }
	else if (moves_over < 30) { a = 9; b = 2; c = 1; d = 9; e = 6; }
	else if (moves_over < 40) { a = 9; b = 2; c = 1; d = 9; e = 6; }
	else if (moves_over < 48) { a = 9; b=1.7; c = 0; d = 8; e = 5; }
	else if (moves_over < 53) { a = 9; b=1.2; c = 0; d = 7; e = 4; }
	else 							  { a = 9; b = 0; c = 0; d = 0; e = 0; }

	int t1 = strat_3(who), t2 = strat_3(opp(who));
	int t3 = no_of_pieces(who), t4 = no_of_pieces(opp(who));

	if ( t1==0 && t2==0)					//game has ended
		{
		if (t3>t4)
			return 100000 + (t3-t4);
		else
			return -100000 - (t4-t3);
		}

	switch(eval_level)
		{
		case 1:	b = c = d = e = 0; break;		//only a i.e difference
		case 2:  d = e = 0; break;				//a,b,c i.e edge & corner & diff
		case 3:  c = 0;				//diff, corner, mobility, pot. mobility
		}

	if (a!=0) sum1 = t3 - t4;
	if (b!=0) sum2 = strat_1(who) - strat_1(opp(who));
	if (c!=0) sum3 = strat_2(who) - strat_2(opp(who));
	if (d!=0) sum4 = t1 - t2;
	if (e!=0) sum5 = strat_4(who) - strat_4(opp(who));

	sum = a*sum1 + b*sum2+ c*sum3 + d*sum4 + e*sum5;
	return sum;
	}

int TheMaster2::strat_1(id who)  //corner discs COMBINED with CONSECUTIVE
	{                              //edge discs give more power
	static int starts[4] = { 0, 7, 56, 63 };
	static int incs[8] = { 1, 8, -1, 8, 1, -8, -1, -8};
	int i,j;
	int corner;
	int to_return=0;

	for(corner=0; corner<=3; corner++)
		{
		int cell = starts[corner];

		if (board[cell] != who) continue;

		to_return += 100;

		for(i=1; i<=6; i++)
			{
			j = cell + i*incs[corner*2];
			if (board[j] != who) break;
			to_return += (i*8);
			}
		for(i=1; i<=6; i++)
			{
			j = cell + i*incs[corner*2+1];
			if (board[j] != who) break;
			to_return += (i*8);
			}
		}

	return to_return;

	}

int TheMaster2::strat_2(id who)  //count of discs on edges
	{
	int i;
	int to_return=0;
	for(i=0; i<=7; i++) to_return += ( (board[i]==who)? 1 : 0);
	for(i=56; i<=63; i++) to_return += ( (board[i]==who)? 1 : 0);
	for(i=8; i<=48; i+=8) to_return += ( (board[i]==who)? 1 : 0);
	for(i=15; i<=55; i++) to_return += ( (board[i]==who)? 1 : 0);
	return to_return;
	}

int TheMaster2::strat_3(id who) //mobility
	{
	return no_of_legal_moves(who);
	}

int TheMaster2::strat_4(id who) //potential mobility
	{
	int to_return=0;
	id you = opp(who);

	for(int i=0; i<=63; i++)
		if (board[i] == 0)
			{
			int r = i/8, c = i%8;
			if (r<=5) to_return += ( (board[i+8]==you)? 1 : 0);
			if (r>=2) to_return += ( (board[i-8]==you)? 1 : 0);
			if (c<=5) to_return += ( (board[i+1]==you)? 1 : 0);
			if (c>=2) to_return += ( (board[i-1]==you)? 1 : 0);
			if (r<=5 && c<=5) to_return += ( (board[i+9]==you)? 1 : 0);
			if (r<=5 && c>=2) to_return += ( (board[i+7]==you)? 1 : 0);
			if (r>=2 && c<=5) to_return += ( (board[i-7]==you)? 1 : 0);
			if (r>=2 && c>=2) to_return += ( (board[i-9]==you)? 1 : 0);
			}

	return to_return;
	}

